/*********************************************************************
 * Desktop Station main Sketch for DSair Hardware Platform
 *
 * Copyright (C) 2020 DesktopStation Co.,Ltd. / Yaasan
 * https://desktopstation.net/
 */

//#define DEBUG

// include the library code:
#include <SPI.h>
#include <avr/wdt.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Functions.h"
#include "FlashAirSharedMem.h"
#include "TrackReporterS88_DS.h"
#include "DSLCDManager.h"
#include "DSSequence.h"
#include "OzOLED.h"


#define FIRMWARE_VER   '5' /* 0-9, a-z, A-Z*/

#define MAX_S88DECODER 1

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2
#define LEDSTATE_NOSD 3
#define LEDSTATE_ANALOG 4


#define SMEM_POWER		0
#define SMEM_ERROR		1
#define SMEM_EDC		2
#define SMEM_CUR		3
#define SMEM_SEQ		4
#define SMEM_S88_1		5
#define SMEM_S88_2		6
#define SMEM_S88_3		7
#define SMEM_S88_4		8

#define SHAREMEM_SIZE 48
#define SHAREMEM_STATUS_SIZE 264 //32 + 16 + 32x2 + 19 * 8= 112 + 144=264
#define SHAREMEM_ACCSIZE 32

#define THRESHOLD_EDC 245  // 1024d * (1.2 / 8) * 9V / 5V
//#define THRESHOLD_OV  553  // 1024d * (1.2 / 8) * 20V / 5V
#define THRESHOLD_OV  768  // 1024d * (1.2 / 8) * 22V / 5V

#define HW_DSS1		0
#define HW_DSAIR1	1
#define HW_DSAIR2	2

#define CMDMODE_WIFI 0
#define CMDMODE_SERIAL 1

#define TIMEOUT_SERIAL 10

#define MAX_LOCSTATUS_WEB	8

#ifdef __LGT8FX8P__
#define ADC_SHIFT 2
#else
#define ADC_SHIFT 0
#endif

typedef struct {
	byte mOV:1;
	byte mOC:1;
	byte mLV:1;
	byte mUnk1:1;
	byte mUnk2:1;
	byte reserved:3;
} DATA_STATUS;

typedef struct {
	word mAddress;//Marklin Address Format
	byte mSpeed;//(1/4 Data)
	byte mDir;
	long mFuncBuf;
} DATA_LOC;


byte gHardware = HW_DSS1;
byte gCMDMode = CMDMODE_WIFI;
uint8_t buffer_sharedStatusMem[SHAREMEM_STATUS_SIZE + 1];
uint8_t buffer_sharedMem[SHAREMEM_SIZE];
uint8_t buffer_lastcmd[SHAREMEM_SIZE];
uint8_t buffer_accessories[SHAREMEM_ACCSIZE];

/* 機関車データ確認 */
DATA_LOC gLocStatus[MAX_LOCSTATUS_WEB];

/* FlashAir接続確認 */
byte gFlag_FlashAir;
uint8_t gFlashAirAccess = 0;

/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);
DSSequence Sequence;

/* OLED */
OzOLED LCD;
DSLCDManager LCDManager;
byte gOLEDMSG_timer = 0;
char gTempText[16];


/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Sensor */
short gOffset_Current = 0;
word gEdc_data = 0;
short gCur_data = 0;
uint8_t gSwSensor = 0;
uint8_t gMaxS88Num = 0;
unsigned short gCurrent_val;
unsigned short gEdc_val;

//Error check
unsigned char gErrorFlag = 0;
unsigned short gOCLevel = 450; //4.5Amax

/* Error Threshold */
word gThreshold_OV = 0;
word gThreshold_LV = 0;
byte gLastErrorNo = 0;
uint8_t gAliveSeq = 0;

/* Parser */
String gRequest;
String function;
word arguments[4];
byte numOfArguments;
word gLocAddresses[4] = {0,0,0,0};
byte numOfLocAddr = 0;
boolean result;

DATA_STATUS gStatus;


//Task Schedule
unsigned long gPreviousL7 = 0; // 1000ms interval(Packet Task)
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval

void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
boolean SDIO_ReadSharedMem();
boolean SDIO_WriteSharedMem_Status();
boolean SDIO_WriteSharedMem_CVReply(word inCVNo, uint8_t inValue);
void SDIO_ClearSharedMem();
int8_t SharedMemRead2(uint16_t adr, uint16_t len, uint8_t buf[]);
int8_t SharedMemWrite2(uint16_t adr, uint16_t len, uint8_t buf[]);


void LOCMNG_Clear(void);
byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd);
byte LOCMNG_SetLocoDir(word inAddr, byte inDir);
byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower);
void LOCMNG_UpdateAcc(word inNo);

void outputDammyPulse();
void Beep(word inLength);

void taskLED(void);
void changeLED(void);

boolean parse();
boolean dispatch();

void Reply200(void);
void Reply300(void);
void Reply301(void);
void Reply302(void);
void Reply304(void);
void Reply303(void);
void ScanS88(void);
void PowerOffByErr(uint8_t inErrNo);
void ReplyPowerPacket(byte inPower);

void addOLEDmessage(const char *String);
void clearOLEDmessage();
void delOLEDmessage();

void PowerOnOffControl(word inOnoff);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	
	/* OLED init */
	LCD.OutputDammy = outputDammyPulse;
	LCD.init();
	LCD.clearDisplay();           //clear the screen and set start position to top left corner
	LCD.setNormalDisplay();       //Set display to Normal mode
	LCD.setPageMode();            //Set addressing mode to Page Mode
	LCD.printString("DSmain R6", 3, 2);
	LCD.printString("Version 0", 3, 3);
	LCD.printString("Welcome onboard", 1, 6);
	
	//LCD Manager
	LCDManager.PrintString = printString;
	LCDManager.PrintStringNOT = printStringNOT;
	LCDManager.PrintStringBig = printStringBig;
	LCDManager.ClearDisplay = clearDisplay;
	LCDManager.DrawIcon = drawIcon;
	LCDManager.DrawSpeedBar = drawSpeedBar;

	/* Debug message for PC */
	Serial.begin(115200);
	while (!Serial);
	
	Serial.println("Desktop Station air2 XL");

	//初期化
	LOCMNG_Clear();
	
	for( int i = 0; i < SHAREMEM_ACCSIZE; i++)
	{
		buffer_accessories[i] = 0;
	}
	
	//Pin mode
	pinMode(PIN_EDC, INPUT);//Edc
	pinMode(SENSOR_CURRENT, INPUT);
	pinMode(PIN_RUNLED, OUTPUT);
	pinMode(PIN_ERRLED, INPUT);
	pinMode(PIN_SPK, OUTPUT);
	pinMode(PIN_PWMA, OUTPUT);
	pinMode(PIN_PWMB, OUTPUT);
	
	
	/* DSCore初期化 */
	DSCore.Init();
	
	/* 線路に電源遮断 */
	DS_Power(0);
	
	/* Dammy sleep */
	CalcCurrentOffset();
	
	/* 各種クラスの初期化 */
	Sequence.Init();
	
	
	//Hardware seriesのチェック
	gHardware = HW_DSAIR2;
	gThreshold_OV = THRESHOLD_OV;
	gThreshold_LV = THRESHOLD_EDC;
	
	/* 共有メモリクリア */
	ClearShareMem();
	
	Serial.println(F("HW: DSair2XL"));
	Serial.println(F("-------------"));
	
	// Initialize SD card.
	Serial.print(F("Init SD card..."));  
	
	if (SharedMemInit() >= 0) 
	{
		Serial.println(F("OK"));
		gFlag_FlashAir = 1;
		
		Beep(500);
		
		//共有メモリをクリア
		SDIO_ClearSharedMem();
		SDIO_WriteSharedMem_Status();
		
	} else {
		
		Serial.println(F("Retry"));
		
		/* Retry */
		
		if (SharedMemInit() >= 0) 
		{
			Serial.println(F("OK"));
			gFlag_FlashAir = 1;
			
			Beep(1000);
			
			//共有メモリをクリア
			SDIO_ClearSharedMem();
			SDIO_WriteSharedMem_Status();
		}
		else
		{
			/* 失敗 */
			Serial.println(F("NG"));
			gFlag_FlashAir = 0;
			
			Beep(2000);
		}
	}
	
	/* LED turn on */
	changeLED();
	
	Serial.println(F("100 Ready"));
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	gPreviousL7 = millis();
	
	//OLED Update
	clearDisplay();
	
	if( gFlag_FlashAir == 1 )
	{
		LCD.printString("FlashAir", 0, 0);
	}
	else
	{
		LCD.printString("USB", 0, 0);
	}
	
	//POWER OFF
	drawIcon(15, 0, 7, false);
	
	/* ウォッチドッグ有効(4sec) */
	wdt_enable(10); // WDT reset
	
	//ENABLE INTERRUPTION
	sei();
	
}


boolean parse() 
{
	int aPosSlash = -1;

	char aTemp[SHAREMEM_SIZE];
	gRequest.toCharArray(aTemp, SHAREMEM_SIZE);

	int lpar = gRequest.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(gRequest.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int aAdddrOffset = 0;
	int comma = gRequest.indexOf(',', offset);
	numOfArguments = 0;
	numOfLocAddr = 0;
	
	while (comma != -1) 
	{
		String tmp = gRequest.substring(offset, comma);
		tmp.trim();
		
		/* スラッシュを探索する */
		aPosSlash = tmp.indexOf('/', 0);
		
		if( aPosSlash > 0)
		{
			tmp = tmp + "/";
			
			//Dammy値を指定
			arguments[numOfArguments++] = 0;
			aAdddrOffset = 0;
			
			//スラッシュはアドレス部だけに許可
			while( aPosSlash != -1)
			{
				String tmp2 = tmp.substring(aAdddrOffset, aPosSlash);
				tmp2.trim();
				
				if( numOfLocAddr < 4)
				{
					gLocAddresses[numOfLocAddr] = stringToWord(tmp2);
					numOfLocAddr++;
				}
				
				aAdddrOffset = aPosSlash + 1;
				aPosSlash = tmp.indexOf('/', aAdddrOffset);
			}
			
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
		else
		{
			arguments[numOfArguments++] = stringToWord(tmp);
			offset = comma + 1;
			comma = gRequest.indexOf(',', offset);
		}
	}

	int rpar = gRequest.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = gRequest.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	boolean aResult;
	byte aDir = 0;
	
	if ((function == "setLocoDirection") || ((function == "DI")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoDir(gLocAddresses[i], aDir);
				aResult =  DSCore.SetLocoDirection(gLocAddresses[i], aDir);
			}
			
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoDir(arguments[0], aDir);
			aResult =  DSCore.SetLocoDirection(arguments[0], aDir);
		}
		
		return aResult;
	}
	else if ((function == "setLocoFunction") || ((function == "FN")))
	{
		LOCMNG_SetLocoFunc(arguments[0], arguments[1], arguments[2]);
		
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "TO")))
	{
		//テーブル更新
		//MM2: 1(0x3000) - 320(0x3140)
		//DCC: 1(0x3800) - 2044(0x3FFC)
		
		word aAccAddr = 0;
		
		if( (arguments[0] >= 0x3000) && (arguments[0] <= 0x3140))
		{
			aAccAddr = arguments[0] - 0x3000;
		}
		else if( (arguments[0] >= 0x3800) && (arguments[0] <= 0x3FFC))
		{
			aAccAddr = arguments[0] - 0x3800;
		}
		
		if( (aAccAddr >> 3) <= 31)
		{
			if( arguments[1] == 0)
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] & ~(1UL << (aAccAddr & 0x07));
			}
			else
			{
				buffer_accessories[aAccAddr >> 3] = buffer_accessories[aAccAddr >> 3] | (1UL << (aAccAddr & 0x07));
			}
			
			LOCMNG_UpdateAcc(aAccAddr >> 3);
		}
		
		//ポイントパケット出力登録
		
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "PW")))
	{
		PowerOnOffControl(arguments[0]);
		
		/* 線路電源処理 */
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "SP")))
	{
		if( numOfLocAddr > 0)
		{
			//重連対応
			for( int i = 0; i < numOfLocAddr; i++)
			{
				LOCMNG_SetLocoSpeed(gLocAddresses[i], arguments[1]);
				aResult =  DSCore.SetLocoSpeedEx(gLocAddresses[i], arguments[1], arguments[2]);
			}
			aResult = true;
		}
		else
		{
			//非重連
			LOCMNG_SetLocoSpeed(arguments[0], arguments[1]);
			aResult =  DSCore.SetLocoSpeedEx(arguments[0], arguments[1], arguments[2]);
		}
		
		return aResult;
	}
	else if (function == "DC")
	{
		
		return true;
	}
	else if ((function == "reset") || ((function == "RS")))
	{
		#ifdef DEBUG
		Serial.println(F("100 Ready"));
		#endif
		
		/* リセットパケットを投げる */
		if( gLED_State == LEDSTATE_RUN)
		{
			//内部のデータバッファを削除する
			DSCore.Clear();
			
			//リセット
			DSCore.SendReset();
			
		}
		
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "PG")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		Serial.println(F("@DSG,00,03,00,02,00,00,00,00"));
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "GV")))
	{
		
		/* Clear Share Mem*/
		//SDIO_ClearSharedMem();
		
		//Clear last CV data
		SDIO_WriteSharedMem_CVReply(0, 0);
		
		//Send to FlashAir Shared Memory
		SDIO_WriteSharedMem_Status();
		
		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);
		
		/* Reset watchdog timer. */
		wdt_reset(); 

		aValue = 0;
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 0, 0);
		
		
		//Send to Serial
		Serial.print("@CV,");
		Serial.print(arguments[0]);
		Serial.print(",");
		Serial.print(arguments[1]);
		Serial.print(",");
		
		if( aResult == true)
		{
			Serial.print(aValue);
			
			//Write to FlashAir
			SDIO_WriteSharedMem_CVReply(arguments[1], aValue);
			SDIO_WriteSharedMem_Status();
		}
		else
		{
			Serial.print(0xFFFF);
			
			//Write to FlashAir
			SDIO_WriteSharedMem_CVReply(0, 255);//255 is error no (Read. but not found)
			SDIO_WriteSharedMem_Status();
		}
		
		Serial.println(",");
		
		delOLEDmessage();
		sprintf(gTempText, "CV read No.%d=%d",  arguments[1], aValue);
		addOLEDmessage(gTempText);
		
		/* LED turn off */
		changeLED();
		
		//FlashAir 共有メモリSPI送信
		SDIO_WriteSharedMem_Status();
		
		
		return true;
	}
	else if (function == "gS8")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		//S88 Normal mode
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		if( aMaxS88Num > 2)
		{
			aMaxS88Num = 2;
		}
		
		gMaxS88Num = aMaxS88Num;
		
		char aStrTemp[2];
		sprintf(aStrTemp, "%d", gMaxS88Num);
		buffer_sharedStatusMem[20] = aStrTemp[0];
		
		return true;
		
	} /* getS88 for Wi-Fi */
	else if (function == "getS88")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		Serial.print("@S88,");

		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  Serial.print((word)aFlags, HEX);
		  Serial.print(",");
		}
			
		Serial.println("");
		
		return true;
		
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "SV")))
	{
		
		Serial.print("@CVWrite,");
		Serial.print(arguments[1]);
		Serial.print(",");
		Serial.print(arguments[2]);
		Serial.println("");
		
		
		//Clear last CV data
		SDIO_WriteSharedMem_CVReply(0, 0);
		
		//Send to FlashAir Shared Memory
		SDIO_WriteSharedMem_Status();
		
		/* LED turn on */
		digitalWrite(PIN_RUNLED, HIGH);
		
		/* Reset watchdog timer. */
		wdt_reset(); 
		
		DSCore.WriteConfig_Dir( arguments[1], arguments[2]);
		
		//Write to FlashAir
		SDIO_WriteSharedMem_CVReply(arguments[1], arguments[2]);
		SDIO_WriteSharedMem_Status();
		
	
		/* LED turn off */
		changeLED();

		return true;
	}
	else if ((function == "cutout") || ((function == "CO")))
	{
		/* Railcom CutOut */
		DSCore.gCutOut = (arguments[0] == 1) ? 1: 0;
		
		return true;
	}

	else
	{
		return false;
	}
}

/***********************************************************************
* 
* Timer Interval
* 
************************************************************************/

void loop()
{
	long aEdc;
	signed long aCur;
	uint8_t aErrNo = 1;
	
	
	//パルスジェネレータのスキャン
	DSCore.Scan();
	
	if( (millis() - gPreviousL7) >= 200)
	{
		//Reset task
		gPreviousL7 = millis();
		
		//Sequence
		if( Sequence.Interval() == true)
		{
			
			switch(Sequence.gButtonStatus)
			{
			case SQBUTTON_RUN:
				
				byte aOnOff = (gLED_State == LEDSTATE_STOP) ? 1 : 0;
				
				PowerOnOffControl(aOnOff);
				DSCore.SetPower(aOnOff);
				break;
				
			}
			
		}
	}

	
	if( (millis() - gPreviousL5) >= 200)
	{
		//Reset task
		gPreviousL5 = millis();
		
		
		//LED Control
		taskLED();
		LCDManager.Interval();
		
		//Message Control
		clearOLEDmessage();
	}
	
	if( (millis() - gPreviousL8) >= 500)
	{
		
		//Reset task
		gPreviousL8 = millis();
		
		//S88 Scan
		ScanS88();
		
		/* Sensor check */
		if( gSwSensor == 0)
		{
			gSwSensor = 1;
			gEdc_data = analogRead(PIN_EDC);
			
			/* OV check */
			if( gEdc_data > gThreshold_OV )
			{
				aErrNo = 4;
			}
			
			/* LV check */
			if( gEdc_data < gThreshold_LV )
			{
				aErrNo = 2;
				
			}
			
			aEdc = gEdc_data >> 1;//512に分解能を落とす
			aEdc = ((aEdc * 126) / 3) >> 6;
			
			if( aEdc > 255)
			{
				aEdc = 255;
			}
			
			gEdc_val = aEdc;
			
			//Sensor Write
			RegisterShareMem(SMEM_EDC, (uint8_t)aEdc);
			
		}
		else
		{
			gSwSensor = 0;
			gCur_data = analogRead(PIN_CURRENT) >> ADC_SHIFT;
			
			/* 電流の向きを正規化 */
			if( gCur_data >= gOffset_Current)
			{
				gCur_data = gCur_data - gOffset_Current;
			}
			else
			{
				gCur_data = gOffset_Current - gCur_data;
			}			
			
			aCur = ((long)gCur_data * 651) / 213;
			
			if( aCur > 1250)
			{
				aCur = 1250;
			}
			
			gCurrent_val = aCur;
			
			RegisterShareMem(SMEM_CUR, (uint8_t)aCur);
			
		}
		
		
		/* Hardware Protection Check  */
		
		if( gCurrent_val >= gOCLevel)
		{
			//Over Current (OC)
			aErrNo = 8;
		}
		
		
		/* エラー処理 */
		if( aErrNo > 1)
		{
			if( DSCore.IsPower() == 1)
			{
				
				PowerOffByErr(aErrNo);
				
				//OLED表示
				delOLEDmessage();
				switch(aErrNo)
				{
				case 2: 
					addOLEDmessage("Err: LV");
					break;
				case 4: 
					addOLEDmessage("Err: OV");
					break;
				case 8: 
					addOLEDmessage("Err: OC");
					break;
				default:
					addOLEDmessage("Err: UNK");
					break;
				
				}
				
				
			}
			
			//FlashAirのバッファにセット
			RegisterShareMem(SMEM_ERROR, (uint8_t)aErrNo);
			
			
		}
		
		//送信回数
		RegisterShareMem(SMEM_SEQ, (uint8_t)gAliveSeq);
		
		if( gAliveSeq >= 99)
		{
			gAliveSeq = 0;
		}
		else
		{
			gAliveSeq++;
		}
		
		/* 表示 */
		printSensors();
		
		//FlashAir 共有メモリSPI送信
		SDIO_WriteSharedMem_Status();
	}
	
	//FlashAir 有り無しの処理分岐
	if( gFlag_FlashAir != 0)
	{
		gFlashAirAccess++;
		
		if( gFlashAirAccess > 4)
		{
			gFlashAirAccess = 0;
		}
		
		if( gFlashAirAccess == 0)
		{
			
			/* コマンド受信＆処理 */
			boolean aRecvShareMem= SDIO_ReadSharedMem();
			
			if( aRecvShareMem == true)
			{
				//データをシリアルに流し込む
				gRequest = "";
				
				for( int i = 0; i < SHAREMEM_SIZE; i++)
				{
					if( buffer_sharedMem[i] != ')' )
					{
						char aWord = (char)(buffer_sharedMem[i]);
						
						gRequest = gRequest + aWord;
					}
					else
					{
						gRequest = gRequest + ")";
						break;
					}
				}
				
				if (parse()) 
				{
					#ifdef DEBUG
					Serial.println(gRequest);
					#endif
					
					if (dispatch()) 
					{
						#ifdef DEBUG
						Reply200();
						#endif
					}
					else 
					{
						#ifdef DEBUG
						Reply300();
						#endif
					}
				} 
				else 
				{
					#ifdef DEBUG
					Reply301();
					#endif
				}
				
			}
			else
			{
				/* Nothing to do */
				
				
			}
		}
	}
	
	

	int aReceived = receiveRequest();
	
	if( aReceived > 0)
	{
		gFlag_FlashAir = 0;
		
  		if (parse()) {
			
  			#ifdef DEBUG
			Serial.println(gRequest);
			#endif
  			
			if (dispatch()) {
				Reply200();
			} else {
				Reply300();
			}
  			
		} else {
			Reply301();
		}
		
		/* Reply to Desktop Station */
	}
	else
	{
		/* Nothing to do */
		
	}
		

	/* Reset watchdog timer. */
	wdt_reset(); 

}

void Reply200(void)
{
	Serial.println(F("200 Ok"));
}

void Reply300(void)
{
	Serial.println(F("300 Cmd err"));
}

void Reply301(void)
{
	Serial.println(F("301 Stx err"));
}

void Reply302(void)
{
	Serial.println(F("302 Timeout"));
}

void Reply303(void)
{
	Serial.println(F("303 Unkn err"));
}

void Reply304(void)
{
	Serial.println(F("304 Ser err"));
}


void ScanS88(void)
{
	
	if( gFlag_FlashAir == 0)
	{
		return;
	}

	if( gMaxS88Num > 0)
	{
		reporter.refresh(gMaxS88Num);

		for( int j = 0; j < gMaxS88Num; j++)
		{
			RegisterShareMem(SMEM_S88_1 + j * 2, reporter.getByte(j << 1));
			RegisterShareMem(SMEM_S88_2 + j * 2, reporter.getByte((j << 1) + 1));
			
		}
		
		
	}
}



void PowerOffByErr(uint8_t inErrNo)
{
	DSCore.SetPower(0);
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
	gLED_Counter =0;
	gLastErrorNo = inErrNo;
	
	changeLED();
	
	
	Beep(3000);
	
	/* FlashAirにも反映 */
	RegisterShareMem(SMEM_POWER, 0);
	
	//POWER OFF
	drawIcon(15, 0, 7, false);
	
	//Reply
	ReplyPowerPacket(0);

	
}

void ReplyPowerPacket(byte inPower)
{
	#ifdef DEBUG
	/* Power */
	Serial.print("@PWR,");
	Serial.print(inPower);
	Serial.println(",");
	#endif
}


void changeLED(void)
{
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, HIGH);
		digitalWrite(PIN_ERRLED, HIGH);
		break;
	
	case LEDSTATE_ANALOG:
		break;
		
	case LEDSTATE_ERR:
		digitalWrite(PIN_RUNLED, LOW);
		pinMode(PIN_ERRLED, OUTPUT);
		digitalWrite(PIN_ERRLED, LOW);
		break;
		
	case LEDSTATE_NOSD:
		pinMode(PIN_ERRLED, OUTPUT);
		break;
	}
}

void taskLED(void)
{
	
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	
	if( gLED_Counter >= 16)
	{
		gLED_Counter = 0;
	}
	
	//LED操作
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		
		if( gLED_Toggle == 0)
		{
			digitalWrite(PIN_ERRLED, HIGH);
		}
		else
		{
			digitalWrite(PIN_ERRLED, LOW);
		}
		
		gLED_Toggle = (~gLED_Toggle) & 1;
		break;
	
	case LEDSTATE_ANALOG:
		
		
		break;
		
	case LEDSTATE_NOSD:
		
		if( gLED_Counter <= 2)
		{
			digitalWrite(PIN_ERRLED, HIGH);
		}
		else
		{
			digitalWrite(PIN_ERRLED, LOW);
		}
		
		break;
	default:
		break;
	}
}


void SDIO_ClearSharedMem()
{	
	//Clear Buffer
	memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	if( SharedMemWrite2( 0, SHAREMEM_SIZE, buffer_sharedMem))
	{
		Serial.println(F("\nwrite err"));
		return;
	}	
	
	//Update LastCommandBuffer
	for( byte i = 0; i < SHAREMEM_SIZE; i++)
	{
		buffer_lastcmd[i] = 0;
	}
	
}

boolean SDIO_ReadSharedMem()
{
	boolean aResultFnc = false;
	uint8_t aChk = 0;
	
	if( gFlag_FlashAir == 0)
	{
		return aResultFnc;
	}
	
	//Clear Buffer
	//memset(buffer_sharedMem, 0, SHAREMEM_SIZE);
	
	if( SharedMemRead2(0, SHAREMEM_SIZE, buffer_sharedMem))
	{
		return false;
	}
	
	//Check between received command and last one.
	for( byte i = 0; i < SHAREMEM_SIZE; i++)
	{
		if( buffer_sharedMem[i] == 0x00)
		{
			break;
		}
		
		if( buffer_lastcmd[i] != buffer_sharedMem[i])
		{
			aChk++;
			break;
		}
	}
	
	if( buffer_sharedMem[0] == 0x00)
	{
		//Shared memory is empty
		aResultFnc = false;
	}
	else if( aChk == 0)
	{
		//Same command received.
		aResultFnc = false;
	}
	else
	{
		aResultFnc = true;
		
		//Update LastCommandBuffer
		for( byte i = 0; i < SHAREMEM_SIZE; i++)
		{
			buffer_lastcmd[i] = buffer_sharedMem[i];
		}
		
	}
	
	return aResultFnc;
}


boolean SDIO_WriteSharedMem_Status()
{
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	if( SharedMemWrite2( 0x80, SHAREMEM_STATUS_SIZE, buffer_sharedStatusMem))
	{
		Serial.println(F("\nwrite err"));
		
		return false;
	}
	else
	{
		return true;
	}
}

void ClearShareMem()
{
	
	buffer_sharedStatusMem[0] = 0x4e;//PWR
	buffer_sharedStatusMem[1] = 0x2c;//カンマ
	buffer_sharedStatusMem[2] = 0x30;//ERR
	buffer_sharedStatusMem[3] = 0x2c;//カンマ
	buffer_sharedStatusMem[4] = FIRMWARE_VER;//VER
	buffer_sharedStatusMem[5] = 0x2c;//カンマ
	buffer_sharedStatusMem[6] = 0x30;//制御車両数(0-15, 16進数)
	buffer_sharedStatusMem[7] = 0x2c;//カンマ
	buffer_sharedStatusMem[8] = 0x30;//電圧
	buffer_sharedStatusMem[9] = 0x30;
	buffer_sharedStatusMem[10] = 0x30;
	buffer_sharedStatusMem[11] = 0x2c;//カンマ
	buffer_sharedStatusMem[12] = 0x30;//電流
	buffer_sharedStatusMem[13] = 0x30;
	buffer_sharedStatusMem[14] = 0x2c;//カンマ
	buffer_sharedStatusMem[15] = 0x30 + gHardware;//HW version
	buffer_sharedStatusMem[16] = 0x2c;//カンマ
	buffer_sharedStatusMem[17] = 0x30;//送信回数
	buffer_sharedStatusMem[18] = 0x30;//送信回数（生存チェック)
	buffer_sharedStatusMem[19] = 0x2c;//カンマ
	buffer_sharedStatusMem[20] = 0x30;//S88 MAX
	buffer_sharedStatusMem[21] = 0x30;//S88 1L
	buffer_sharedStatusMem[22] = 0x30;//S88 1H
	buffer_sharedStatusMem[23] = 0x30;//S88 2L
	buffer_sharedStatusMem[24] = 0x30;//S88 2H
	buffer_sharedStatusMem[25] = 0x30;//S88 3L
	buffer_sharedStatusMem[26] = 0x30;//S88 3H
	buffer_sharedStatusMem[27] = 0x30;//S88 4L
	buffer_sharedStatusMem[28] = 0x30;//S88 4H
	buffer_sharedStatusMem[29] = 0x30;//S88 reserved
	buffer_sharedStatusMem[30] = 0x30;//S88 reserved
	buffer_sharedStatusMem[31] = 0x3b;
	
	//CV
	for(int i = 0; i < 15; i++)
	{
		buffer_sharedStatusMem[32 + i] = 0x30;
	}
	
	buffer_sharedStatusMem[47] = 0x3b;
	
	
	//ポイント
	
	for(int i = 0; i < SHAREMEM_ACCSIZE; i++)
	{
		buffer_sharedStatusMem[48 + i * 2] = 0x30;
		buffer_sharedStatusMem[48 + i * 2 + 1] = 0x30;
	}
	
		buffer_sharedStatusMem[112] = 0x3b;	//;
	
	
	//アクティブな車両データ(最新5つ)
	
	for(int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		buffer_sharedStatusMem[113 + i * 19] = 0x30;	//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 1] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 2] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 3] = 0x30;//LocAddr
		buffer_sharedStatusMem[113 + i * 19 + 4] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 5] = 0x30;//Spd
		buffer_sharedStatusMem[113 + i * 19 + 6] = 0x30;//Spd
		buffer_sharedStatusMem[113 + i * 19 + 7] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 8] = 0x30;//Dir
		buffer_sharedStatusMem[113 + i * 19 + 9] = 0x2C;//,
		buffer_sharedStatusMem[113 + i * 19 + 10] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 11] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 12] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 13] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 14] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 15] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 16] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 17] = 0x30;//Fnc
		buffer_sharedStatusMem[113 + i * 19 + 18] = 0x2f;///
	}	
	
	
}

void RegisterShareMem(uint8_t inNumber, uint8_t inValue)
{
	/*
	0x1000-0x107F FlashAir→Arduino コマンド
	0x1080-0x10AF ステータス
	0x10B0-0x10C0 CV Reply
	*/
	
	char aTxtBuf[5];
	
	if( gFlag_FlashAir == 0)
	{
		return;
	}
	
	
	switch(inNumber)
	{
	case SMEM_POWER:
		//Power
		buffer_sharedStatusMem[0] = (inValue == 1) ? 0x59 : 0x4e;
		break;
		
	case SMEM_ERROR:
		//ErrorCode
		sprintf(aTxtBuf,"%d",inValue);//0x00は無効
		buffer_sharedStatusMem[2] = aTxtBuf[0];
		break;
		
	case SMEM_EDC:
		//電圧
		sprintf(aTxtBuf,"%03d",inValue);//0x00は無効
		buffer_sharedStatusMem[8] = aTxtBuf[0];
		buffer_sharedStatusMem[9] = aTxtBuf[1];
		buffer_sharedStatusMem[10] = aTxtBuf[2];
		
		break;
		
	case SMEM_CUR:
		//電流
		sprintf(aTxtBuf,"%03d",inValue);//0x00は無効
		buffer_sharedStatusMem[12] = aTxtBuf[0];
		buffer_sharedStatusMem[13] = aTxtBuf[1];
		break;
		
	case SMEM_SEQ:
		sprintf(aTxtBuf,"%02d",inValue);//0x00は無効
		buffer_sharedStatusMem[17] = aTxtBuf[0];
		buffer_sharedStatusMem[18] = aTxtBuf[1];
		break;
	case SMEM_S88_1:
		sprintf(aTxtBuf,"%02X",inValue);//0x00は無効
		buffer_sharedStatusMem[21] = aTxtBuf[0];
		buffer_sharedStatusMem[22] = aTxtBuf[1];
		break;
	case SMEM_S88_2:
		sprintf(aTxtBuf,"%02X",inValue);//0x00は無効
		buffer_sharedStatusMem[23] = aTxtBuf[0];
		buffer_sharedStatusMem[24] = aTxtBuf[1];
		break;
	case SMEM_S88_3:
		sprintf(aTxtBuf,"%02X",inValue);//0x00は無効
		buffer_sharedStatusMem[25] = aTxtBuf[0];
		buffer_sharedStatusMem[26] = aTxtBuf[1];
		break;
	case SMEM_S88_4:
		sprintf(aTxtBuf,"%02X",inValue);//0x00は無効
		buffer_sharedStatusMem[27] = aTxtBuf[0];
		buffer_sharedStatusMem[28] = aTxtBuf[1];
		break;
		
	}
}

boolean SDIO_WriteSharedMem_CVReply(word inCVNo, uint8_t inValue)
{
	byte i;
	
	if( gFlag_FlashAir == 0)
	{
		return false;
	}
	
	char aBufCVReply[16];
	
	if( (inCVNo == 0) && (inValue == 0) )
	{
		for( i = 0; i < 15; i++)
		{
			buffer_sharedStatusMem[32 + i] = 0x30;
		}
	}
	else
	{
		
		//@CV,9999,255, (13文字)
		sprintf(aBufCVReply, "@CV,%04d,%03d,  ", inCVNo, inValue);
		
		for( i = 0; i < 15; i++)
		{
			buffer_sharedStatusMem[32 + i] = aBufCVReply[i];
		}	
	}
	
	
	return true;
}


void LOCMNG_UpdateAcc(word inNo)
{
	char aBuf2[2];
	
	sprintf(aBuf2, "%02x", buffer_accessories[inNo]);
	
	buffer_sharedStatusMem[48 + inNo * 2] = aBuf2[1];
	buffer_sharedStatusMem[48 + inNo * 2 + 1] = aBuf2[0];
	
}

//FFFF,1,FFFFFFFF



void CalcCurrentOffset()
{
	unsigned short aCurrentSum = 0;
	
	if( gOffset_Current == 0)
	{
		/* 直流中間に電圧が印加されてないと、実際の動作時のADCのオフセット誤差が異なる */
		if( analogRead(PIN_EDC) >= THRESHOLD_EDC)
		{
			
			for( int i = 0; i < 4; i++)
			{
				aCurrentSum = aCurrentSum + (analogRead(PIN_CURRENT) >> ADC_SHIFT);
				delay(10);
			}
			
			//平均値を算出
			gOffset_Current = aCurrentSum >> 2;
		}
	}
	
	/* 検出できない場合は512をオフセット値にセット */
	if( gOffset_Current == 0)
	{
		gOffset_Current = 512;
	}
}

void LOCMNG_Clear(void)
{
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		gLocStatus[i].mAddress = 0;
		gLocStatus[i].mSpeed = 0;
		gLocStatus[i].mDir = 0;
		gLocStatus[i].mFuncBuf = 0;
	}
}


byte LOCMNG_SetLocoSpeed(word inAddr, word inSpd)
{
	byte aIndex = 255;
	word aSpeed = 0;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		aSpeed = inSpd >> 2;
		
		if( aSpeed > 255)
		{
			aSpeed = 255;
		}
		
		char aStrAddr[5];
		char aStrSpd[3];
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		sprintf(aStrSpd, "%02X", aSpeed);
		
		/* FlashAir共有メモリ書き換え */
		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 5] = aStrSpd[0];//Spd
		buffer_sharedStatusMem[113 + aIndex * 19 + 6] = aStrSpd[1];//Spd
		
	}
	
	return 0;
}

byte LOCMNG_SetLocoDir(word inAddr, byte inDir)
{
	byte aIndex = 255;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		
		if( gLocStatus[aIndex].mDir != inDir)
		{
			gLocStatus[aIndex].mDir = inDir;
			
			buffer_sharedStatusMem[113 + aIndex * 19 + 5] = 0x30;//Spd
			buffer_sharedStatusMem[113 + aIndex * 19 + 6] = 0x30;//Spd
		}
		
		char aStrAddr[5];
		char aStrDir;
		
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);
		aStrDir = (gLocStatus[aIndex].mDir == 0) ? '0' : '1';
		
		/* FlashAir共有メモリ書き換え */
		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 8] = aStrDir;//Dir
	}
	
	return 0;
}

byte LOCMNG_SetLocoFunc(word inAddr, byte inNo, byte inPower)
{
	char aStrAddr[5];
	char aStrFunc1[5];
	char aStrFunc2[5];
	byte aIndex = 255;
	long aFuncBuf;
	
	for( int i = 0; i < MAX_LOCSTATUS_WEB; i++)
	{
		if( (gLocStatus[i].mAddress == inAddr) || ( gLocStatus[i].mAddress == 0 ))
		{
			aIndex = i;
			break;
		}
	}
	
	if( aIndex != 255)
	{
		gLocStatus[aIndex].mAddress = inAddr;
		sprintf(aStrAddr, "%04X", gLocStatus[aIndex].mAddress);

		buffer_sharedStatusMem[113 + aIndex * 19	] = aStrAddr[0];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 1] = aStrAddr[1];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 2] = aStrAddr[2];//LocAddr
		buffer_sharedStatusMem[113 + aIndex * 19 + 3] = aStrAddr[3];//LocAddr

		
		if( inPower == 0)
		{
			gLocStatus[aIndex].mFuncBuf = gLocStatus[aIndex].mFuncBuf & ~(1 << inNo);
		}
		else
		{
			gLocStatus[aIndex].mFuncBuf = gLocStatus[aIndex].mFuncBuf | (1 << inNo);
		}
		
		
		aFuncBuf = gLocStatus[aIndex].mFuncBuf & 0xFF;
		sprintf(aStrFunc1, "%04X", (unsigned int)aFuncBuf);
		aFuncBuf = (gLocStatus[aIndex].mFuncBuf >> 8) & 0xFF;
		sprintf(aStrFunc2, "%04X", (unsigned int)aFuncBuf);
		
		
		/* FlashAir共有メモリ書き換え */
		
		
		buffer_sharedStatusMem[113 + aIndex * 19 + 10] = aStrFunc2[0];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 11] = aStrFunc2[1];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 12] = aStrFunc2[2];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 13] = aStrFunc2[3];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 14] = aStrFunc1[0];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 15] = aStrFunc1[1];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 16] = aStrFunc1[2];//Fnc
		buffer_sharedStatusMem[113 + aIndex * 19 + 17] = aStrFunc1[3];//Fnc
	}
	
	return 0;
}


int receiveRequest() {
	char buffer[64];
	int i = 0;
	char aByte;
	int aResult = -1;
	
	unsigned long time = millis();
	
	/* Check serial buffer */
	if (Serial.available() > 0) {
		
		while (1) {
		
			if (Serial.available() > 0) {
			
				aByte = Serial.read();
				
				// Write to text buf.
				buffer[i] = aByte;
				
				// check last identification of text end.
				if (aByte == '\n') {
					aResult = i;
					break;
				}
				
				// increment
				i = i + 1;
				
				// buf over check
				if( i > 64)
				{
					break;
				}
			}
			else
			{
				/* Check timeout. */
				if( millis() > time + 1000)
				{
					aResult = -1;
					break;
				}
			}
		}
	}
	else
	{
		/* Not received. */
		aResult = 0;
	}
	
	if( aResult > 0)
	{
	
		buffer[i] = '\0';
		
		gRequest = String(buffer);
	}
	else
	{
		gRequest = "";
	}
	
	return aResult;
}

void changedTargetSpeed(const word inAddr, const word inSpeed, const bool inUpdate)
{
	byte aProtocol = 2;
	
	if( inAddr <= 255)
	{
		aProtocol = 0;
	}
	
	if( inUpdate == true)
	{
		DSCore.SetLocoSpeedEx(inAddr, inSpeed, aProtocol);
	}
}

void changedTargetFunc(const word inAddr, const byte inFuncNo, const byte inFuncPower, const bool inUpdate)
{
	if( inUpdate == true)
	{
		DSCore.SetLocoFunction(inAddr, inFuncNo, inFuncPower);
	}

}

void changedTargetDir(const word inAddr, const byte inDir, const bool inUpdate)
{
	
	if( inUpdate == true)
	{
		DSCore.SetLocoDirection(inAddr, inDir);
	}
}

void changedTurnout(const word inAddr, const byte inDir, const bool inUpdate)
{
	if( inUpdate == true)
	{
		DSCore.SetTurnout(inAddr, inDir);
	}
	
}


void changedPowerStatus(const byte inPower)
{
	
	LCDManager.RegisterPower(inPower);
	LCDManager.Interval();
	
	DSCore.SetPower(inPower);
	ReplyPowerPacket(inPower);
	
	if( (gErrorFlag > 0) && (inPower == POWER_ON))
	{
		gErrorFlag = 0;
	}

	
}


//FlashAirアクセス時のオーバーヘッド対策用
int8_t SharedMemRead2(uint16_t adr, uint16_t len, uint8_t buf[])
{
	if(adr + len > 512)
		return -1;
	if(len == 0)
		return -1;
		
	uint32_t sd_adr = 0x1000 + adr;
	uint32_t arg = 0x90000000 | ((sd_adr & 0x1FFFF) << 9) | ((len - 1) & 0x1FF);

	DSCore.TogglePulse();
	
	if(sd_cmd(17,arg) != 0)
	{
		return -2;
	}

	DSCore.TogglePulse();
	
	while((SPI.transfer(0xFF)) != 0xFE);
 
	for (uint16_t i = 0; i < 514; i++) {
		if(i < len){
			buf[i] = SPI.transfer(0xFF);
		}else{
			SPI.transfer(0xFF);
		}
		
		DSCore.TogglePulse();
	}
	cs_release();
	return 0;
}


int8_t SharedMemWrite2(uint16_t adr, uint16_t len, uint8_t buf[])
{
	if(adr + len > 512)
		return -1;
	if(len == 0)
		return -1;
		
	uint32_t sd_adr = 0x1000 + adr;
	uint32_t arg = 0x90000000 | ((sd_adr & 0x1FFFF) << 9) | ((len - 1) & 0x1FF);
	
	DSCore.TogglePulse();
	
	if(sd_cmd(24,arg) != 0)
	{
		return -2;
	}
	
	DSCore.TogglePulse();
	spi_trans(0xFE); //BLOCK START
	
	for (uint16_t i = 0; i < 512; i++) {
		if(i < len){
			spi_trans(buf[i]);
		}else{
			spi_trans(0xFF);
		}
		DSCore.TogglePulse();
	}

	//CRC
	spi_trans(0xFF);
	spi_trans(0xFF);
	
	//Ans
	if((spi_trans(0xFF) & 0x1F) != 0x05) //DATA ACCEPTED
	{
		return -3;
	}
	
	//Wait for Busy
	while(spi_trans(0xFF) == 0x00);

	cs_release();
	return 0;
}


void printString(const char *s, int x, int y)
{
	LCD.printString(s, x, y);
}

void printStringNOT(const char *s, int x, int y)
{
	LCD.printStringNOT(s, x, y);
}


void printStringBig(const char *s, int x, int y)
{
	
	//LCD.setCursorXY(x, y);
	LCD.printBigNumberX2(s, x, y);
}

void drawIcon(byte x, byte y, byte num, byte inNot)
{
	LCD.drawIcon(x, y, num, inNot);
	
}

void drawSpeedBar(byte x, byte y, byte speed, byte inNot)
{
	LCD.drawSpeedBar(x, y, speed, false);
}

void clearDisplay()
{
	LCD.lcd_clear();           //clear the screen and set start position to top left corner

}

void outputDammyPulse()
{
	DSCore.TogglePulse();
}

void Beep(word inLength)
{
	int8_t i;
	
	for( i = 0; i < 50; i++)
	{
		digitalWrite(PIN_SPK, HIGH);
		delayMicroseconds(inLength);
		digitalWrite(PIN_SPK, LOW);
		delayMicroseconds(inLength);
	}
}

void printSensors()
{
	unsigned char aFlagDisplay = gErrorFlag;
	
	unsigned short aCurrent = gCurrent_val;
	unsigned char aVoltage = gEdc_val;
	
	unsigned short aCurrent_H = aCurrent / 100;
	unsigned short aCurrent_L = aCurrent - (aCurrent_H * 100);
	unsigned char aVoltage_H = aVoltage / 10;
	unsigned char aVoltage_L = aVoltage - (aVoltage_H * 10);
	
	//情報表示の時は更新する
	LCDManager.SetSensors(aCurrent_H, aCurrent_L, aVoltage_H, aVoltage_L, aFlagDisplay);
}


void changedMain()
{
	//LCDManager.WriteLocAddr(Sequence.CurrentProtocol + Sequence.NextLocoAddress);
	//LCDManager.WriteAccAddr(Sequence.CurrentACCProtocol + Sequence.NextAccAddress);
	
	
	LCDManager.RegisterMain();
	LCDManager.Interval();
}

void changedInfo()
{
	LCDManager.RegisterInfo();
	LCDManager.Interval();
}


void changedMenu(byte inMenuNo)
{
	LCDManager.RegisterMenu(inMenuNo);
	LCDManager.Interval();
}

void changedConfiguration(byte inMenuNo)
{
	LCDManager.RegisterConfig(inMenuNo);
	LCDManager.Interval();
	
	//OCレベル変更
	updateOCLevel();
}

void updateOCLevel(void)
{
	//gOCLevel = (uint16_t)Sequence.GetOCLVMode() * 100;
	
}

void readCVFunc(const byte inSeq, const word inCVAddr, const byte inCVValue)
{
	byte aValue = 0;
	bool aResult;
	
	if( inSeq == 1)
	{
		LCDManager.RegisterCVRead(10, inCVAddr, 0);
		LCDManager.Interval();
		
		aValue = 0;
		aResult = DSCore.ReadConfigEx(inCVAddr, &aValue, 0);
		
		Serial.print("CV Read: Adr.");
		Serial.print(inCVAddr);
		Serial.print(", Val.");
		Serial.print(aValue);
		Serial.println("");
		
		//Result
		if( aResult != true)
		{
			LCDManager.RegisterCVRead(11, inCVAddr, aValue);
		}
		else
		{
			LCDManager.RegisterCVRead(1, inCVAddr, aValue);
		}
		
		LCDManager.Interval();
	}
	else
	{
		LCDManager.RegisterCVRead(inSeq, inCVAddr, 0);
		LCDManager.Interval();
	}

	
}

void writeCVFunc(const byte inSeq, const word inCVAddr, const byte inCVValue)
{
	long aBaseTime;
	word aCurOffset;
	bool aSuccessWrite = false;
	int aCurrent;
	
	if( inSeq == 2)
	{
		LCDManager.RegisterCVWrite(20, inCVAddr, inCVValue);
		LCDManager.Interval();
		
		
		DSCore.WriteConfig_Dir(inCVAddr, inCVValue);
		
		//電流オフセットを取得
		//Sensor.CalcCurrentOffset();
		aCurOffset = 512;//Sensor.GetCurrentOffset();
		
		//5secのタイムアウト or 電流 60mA待ち
		aBaseTime = millis();
		
		while((millis() - aBaseTime) < 5000)
		{
			aCurrent = (int)analogRead(SENSOR_CURRENT) - (int)aCurOffset;
			
			//Serial.print(aCurrent);
			//Serial.print(",");
			
			if( aCurrent > 10)
			{
				//成功
				aSuccessWrite = true;
				
				break;
			}
			
			
		}
		//delay(2000);
		
		
		LCDManager.RegisterCVWrite(inSeq, aSuccessWrite == true ? inCVAddr : 0xFFFF, inCVValue);
		LCDManager.Interval();
		
	}
	else
	{
		LCDManager.RegisterCVWrite(inSeq, inCVAddr, inCVValue);
		LCDManager.Interval();
	}
	
}

void debugMessage(const byte inPower)
{
	Serial.print("DebugSeq: ");
	Serial.print(inPower);
	Serial.println("");

}

void clearOLEDmessage()
{
	
	if( gOLEDMSG_timer == 1)
	{
		
		/* メッセージ削除 */
		delOLEDmessage();
	}
	else if(gOLEDMSG_timer == 0)
	{
		/* Nothing to do */
		
	}
	else
	{
		//デクリメント
		gOLEDMSG_timer--;
	}
	
}

void delOLEDmessage()
{
	byte i;
	
	for( i = 0; i < 15; i++)
	{
		gTempText[i] = ' ';
	}
	
	gTempText[15] = 0x00;
	
	LCD.printString(gTempText, 0, 2);
	gOLEDMSG_timer = 0;
}

void addOLEDmessage(const char *String)
{
	LCD.printString(String, 0, 2);
	gOLEDMSG_timer = 20;
}


void PowerOnOffControl(word inOnoff)
{
	delOLEDmessage();
	
	if( inOnoff == 1)
	{
		
		//電流オフセット計測
		CalcCurrentOffset();
		
		/* 状態変更 */
		gLED_State = LEDSTATE_RUN;
		drawIcon(15, 0, 1, false);
		
		//FlashAirのバッファにセット
		RegisterShareMem(SMEM_POWER, 1);
		RegisterShareMem(SMEM_ERROR, 1);
		
		Beep(500);
		addOLEDmessage("Power ON");
	}
	else
	{
		/* 状態変更 */
		gLED_State = LEDSTATE_STOP;
		drawIcon(15, 0, 7, false);
		RegisterShareMem(SMEM_POWER, 0);
		
		addOLEDmessage("Power OFF");
		Beep(1000);
	}


	/* LED状態の変更 */
	gLED_Counter = 0;
	changeLED();

	/* OLED電源状態変更 */
	LCDManager.RegisterPower(inOnoff);	
	
}
